__version__ = "0.1.2"
__name__ = "ospark"


from ospark.nn.component.weight import Weight, WeightOperator
from ospark.nn.component.basic_module import ModelObject
from ospark.nn.layers import Layer, normalization, activation
from ospark.utility import weight_initializer
from ospark.nn.component import weight
from ospark.nn.model import Model
from ospark.nn.block import Block
from ospark.nn.cell import Cell


from ospark.nn.layers.self_attention import SelfAttentionLayer
import tensorflow as tf
import numpy as np
from ospark.nn.layers.normalization import Normalization
from typing import NoReturn, Tuple, Optional


class ProbSparseAttentionLayer(SelfAttentionLayer):

    def __init__(self,
                 obj_name: str,
                 embedding_size: int,
                 head_number: int,
                 sample_factor: float,
                 dropout_rate: float,
                 training_phase: Optional[bool]=None,
                 is_training: Optional[bool]=None,
                 use_look_ahead: Optional[bool]=False) -> NoReturn:
        super().__init__(obj_name=obj_name,
                         embedding_size=embedding_size,
                         head_number=head_number,
                         dropout_rate=dropout_rate,
                         is_training=is_training,
                         use_look_ahead=use_look_ahead,
                         training_phase=training_phase)
        self._sample_factor = sample_factor
        self._top_u         = None

    @property
    def sample_factor(self) -> float:
        return self._sample_factor

    @property
    def top_u(self) -> None:
        return self._top_u
    def QKV_process(self,
                    Q_input: tf.Tensor,
                    K_input: tf.Tensor,
                    V_input: tf.Tensor,
                    batch_size: tf.int32) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        Q, K, V = super().QKV_process(Q_input=Q_input, K_input=K_input, V_input=V_input)
        Q_bar = self.sampling(Q, K)
        return Q_bar, K, V

    def sampling(self, Q: tf.Tensor, K: tf.Tensor) -> tf.Tensor:
        batch, head_number, Q_sequence_length, embedding_size = tf.shape(Q)
        _    ,           _, K_sequence_length,              _ = tf.cast(tf.shape(K), dtype=tf.float32)
        self._sequence_length = Q_sequence_length
        sampling_number_U     = tf.cast(tf.math.ceil(self.sample_factor * tf.math.log(K_sequence_length)), dtype=tf.int32)
        sampling_number_u     = tf.cast(tf.math.ceil(self.sample_factor * tf.math.log(tf.cast(Q_sequence_length, dtype=tf.float32))), dtype=tf.int32)
        K_samples             = tf.random.categorical(tf.ones(shape=[1, K_sequence_length]), sampling_number_U)
        K_bar                 = tf.gather(K, K_samples[0], axis=-2)
        sample_score          = tf.matmul(Q, tf.transpose(K_bar, [0, 1, 3, 2]))
        max_mean_measurement  = tf.reduce_max(sample_score, axis=-1) - tf.reduce_mean(sample_score, axis=-1)
        self._top_u           = tf.nn.top_k(max_mean_measurement, sampling_number_u, sorted=False).indices.numpy()
        Q_bar = Q.numpy()[np.arange(batch)[:, None, None],
                          np.arange(head_number)[None, :, None],
                          self.top_u,
                          :]
        return Q_bar

    def attention(self, Q: tf.Tensor, K: tf.Tensor, V: tf.Tensor, mask: tf.Tensor=None) -> tf.Tensor:
        batch, head_number, V_sequence_length, embedding_size = tf.shape(V)
        K = tf.transpose(K, [0, 1, 3, 2])
        scaled_dot_product = tf.matmul(Q, K) / tf.math.sqrt(tf.cast(self.embedding_size, dtype=tf.float32))
        if self.use_look_ahead and mask is not None:
            scaled_dot_product += (tf.cast(tf.math.not_equal(self.look_ahead_mask[self.top_u, :] +
                                                             mask[np.arange(batch)[:, None, None],
                                                                  np.arange(head_number)[None, :, None],
                                                                  self.top_u, :], 0), tf.float32) * -1e9)
        elif mask is not None:
            scaled_dot_product += (mask[np.arange(batch)[:, None, None],
                                        np.arange(head_number)[None, :, None],
                                        self.top_u, :] * -1e9)
        mean_V = self.create_mean_value(V)
        scaled_dot_product = tf.nn.softmax(scaled_dot_product)
        mean_V.numpy()[np.arange(batch)[:, None, None],
                       np.arange(head_number)[None, :, None],
                       self.top_u, :] = tf.matmul(scaled_dot_product, V).numpy()
        concat_output = tf.concat(tf.unstack(mean_V, self.head_number, axis=1), axis=2)
        return concat_output

    def create_mean_value(self, value: tf.Tensor) -> tf.Tensor:
        return tf.tile(tf.reduce_mean(value, axis=-2, keepdims=True), [1, 1, self.sequence_length, 1])


from ospark.nn.block.transformer_block import TransformerEncoderBlock, TransformerDecoderBlock
from ospark.nn.layers.deep_attention import DeepEncoderDecoder, DeepAttentionLayer
from ospark.nn.layers.deep_feed_forward import DeepFeedForwardLayer
from typing import Optional, NoReturn
import tensorflow as tf


class DeepTransformerEncoder(TransformerEncoderBlock):

    def profiling_phase(self, input_data: tf.Tensor, catch_variance: list, mask: Optional[tf.Tensor]=None):
        output, catch_variance = self.assigned.attention.profiling_phase(input_data=input_data,
                                                                         catch_variance=catch_variance,
                                                                         mask=mask)
        output, catch_variance = self.assigned.feedforward.profiling_phase(input_data=input_data,
                                                                           catch_variance=catch_variance)
        return output, catch_variance

    def recover_transformer(self) -> NoReturn:
        self.assigned.attention.back_to_standard()
        self.assigned.feedforward.back_to_standard()


class DeepTransformerDecoder(TransformerDecoderBlock):

    def profiling_phase(self,
                        input_data: tf.Tensor,
                        encoder_output: tf.Tensor,
                        catch_variance: list,
                        encoder_padding_mask: Optional[tf.Tensor]=None,
                        decoder_padding_mask: Optional[tf.Tensor]=None
                        ):
        output, catch_variance = self.assigned.attention.profiling_phase(input_data=input_data,
                                                                         catch_variance=catch_variance,
                                                                         mask=decoder_padding_mask)
        output, catch_variance = self.assigned.encode_decode_attention.profiling_phase(input_data=output,
                                                                                       encoder_output=encoder_output,
                                                                                       catch_variance=catch_variance,
                                                                                       mask=encoder_padding_mask)
        output, catch_variance = self.assigned.feedforward.profiling_phase(input_data=output,
                                                                           catch_variance=catch_variance)
        return output, catch_variance

    def recover_transformer(self) -> NoReturn:
        self.assigned.attention.back_to_standard()
        self.assigned.encode_decode_attention.back_to_standard()
        self.assigned.feedforward.back_to_standard()


def exdeep_encoder_block(obj_name:str,
                         embedding_size: int,
                         head_number: int,
                         scale_rate: int,
                         dropout_rate: float,
                         is_training: Optional[bool] = False
                         ) -> DeepTransformerEncoder:
    attention_layer = DeepAttentionLayer(obj_name="attention",
                                         embedding_size=embedding_size,
                                         head_number=head_number,
                                         dropout_rate=dropout_rate,
                                         is_training=is_training)
    feedforward_layer = DeepFeedForwardLayer(obj_name="feedforward",
                                             embedding_size=embedding_size,
                                             scale_rate=scale_rate,
                                             dropout_rate=dropout_rate,
                                             is_training=is_training)
    block = DeepTransformerEncoder(obj_name=obj_name, attention=attention_layer, feedforward=feedforward_layer)
    return block

def exdeep_decoder_block(obj_name:str,
                         embedding_size: int,
                         head_number: int,
                         scale_rate:int,
                         dropout_rate: float,
                         is_training: Optional[bool] = False
                         ) -> DeepTransformerDecoder:
    attention_layer = DeepAttentionLayer(obj_name="attention",
                                         embedding_size=embedding_size,
                                         head_number=head_number,
                                         use_look_ahead=True,
                                         dropout_rate=dropout_rate,
                                         is_training=is_training)
    encoder_deocder_layer = DeepEncoderDecoder(obj_name="encode_decode_attention",
                                               embedding_size=embedding_size,
                                               head_number=head_number,
                                               dropout_rate=dropout_rate,
                                               is_training=is_training)
    feedforward_layer = DeepFeedForwardLayer(obj_name="feedforward",
                                             embedding_size=embedding_size,
                                             scale_rate=scale_rate,
                                             dropout_rate=dropout_rate,
                                             is_training=is_training)
    block = DeepTransformerDecoder(obj_name=obj_name,
                                   attention=attention_layer,
                                   encode_decode_attention=encoder_deocder_layer,
                                   feedforward=feedforward_layer)
    return block




from __future__ import annotations
from typing import Optional, NoReturn, Callable, List, Type
from ospark.nn.layers.self_attention import SelfAttentionLayer, EncoderDecoderAttentionLayer
from ospark.nn.layers.deep_attention import DeepAttentionLayer, DeepEncoderDecoder
from ospark.nn.layers.feed_forward import FeedForwardLayer
from ospark.nn.layers.deep_feed_forward import DeepFeedForwardLayer
from . import Block
import tensorflow as tf


class TransformerEncoderBlock(Block):

    def __init__(self,
                 obj_name: str,
                 attention: SelfAttentionLayer,
                 feedforward: FeedForwardLayer) -> NoReturn:
        super().__init__(obj_name=obj_name)
        self._attention   = attention
        self._feedforward = feedforward
    
    @property
    def attention(self) -> SelfAttentionLayer:
        return self._attention

    @property
    def feedforward(self) -> FeedForwardLayer:
        return self._feedforward

    @classmethod
    def create_via_class(cls,
                         obj_name: str,
                         embedding_size: int,
                         head_number: int,
                         scale_rate: int,
                         attention_cls: Type[SelfAttentionLayer],
                         feedforward_cls: Type[FeedForwardLayer],
                         dropout_rate: float,
                         is_training: Optional[bool]=None) -> TransformerEncoderBlock:
        return cls(obj_name=obj_name, 
                   attention=attention_cls(obj_name="attention",
                                           embedding_size=embedding_size,
                                           head_number=head_number,
                                           dropout_rate=dropout_rate,
                                           is_training=is_training),
                   feedforward=feedforward_cls(obj_name="feedforward",
                                               embedding_size=embedding_size,
                                               scale_rate=scale_rate,
                                               dropout_rate=dropout_rate,
                                               is_training=is_training))

    def in_creating(self) -> NoReturn:
        self.assign(name="attention", component=self.attention)
        self.assign(name="feedforward", component=self.feedforward)
    
    def pipeline(self, input_data: tf.Tensor, mask: Optional[tf.Tensor]=None) -> tf.Tensor:
        output = self.assigned.attention.pipeline(input_data=input_data, mask=mask)
        output = self.assigned.feedforward.pipeline(input_data=output)
        return output


class TransformerDecoderBlock(Block):

    def __init__(self,
                 obj_name: str,
                 attention: SelfAttentionLayer,
                 encode_decode_attention: EncoderDecoderAttentionLayer,
                 feedforward: FeedForwardLayer) -> NoReturn:
        super().__init__(obj_name)
        self._attention               = attention
        self._encode_decode_attention = encode_decode_attention
        self._feedforward             = feedforward
    
    @property
    def attention(self) -> SelfAttentionLayer:
        return self._attention

    @property
    def encode_decode_attention(self) -> EncoderDecoderAttentionLayer:
        return self._encode_decode_attention

    @property
    def feedforward(self) -> FeedForwardLayer:
        return self._feedforward

    @classmethod
    def create_via_class(cls,
                         obj_name: str,
                         embedding_size: int,
                         head_number: int,
                         scale_rate: int,
                         attention_cls: Type[SelfAttentionLayer],
                         encode_decode_attention_cls: Type[EncoderDecoderAttentionLayer],
                         feedforward_cls: Type[FeedForwardLayer],
                         dropout_rate: float,
                         is_training: Optional[bool]=None) -> TransformerDecoderBlock:
        return cls(obj_name=obj_name, 
                   attention=attention_cls(obj_name="attention",
                                           embedding_size=embedding_size,
                                           head_number=head_number,
                                           use_look_ahead=True,
                                           dropout_rate=dropout_rate,
                                           is_training=is_training),
                   encode_decode_attention=encode_decode_attention_cls(obj_name="encode_decode_attention",
                                                                       embedding_size=embedding_size,
                                                                       head_number=head_number,
                                                                       dropout_rate=dropout_rate,
                                                                       is_training=is_training),
                   feedforward=feedforward_cls(obj_name="feedforward",
                                               embedding_size=embedding_size,
                                               scale_rate=scale_rate,
                                               dropout_rate=dropout_rate,
                                               is_training=is_training))

    def in_creating(self) -> NoReturn:
        self.assign(name="attention", component=self.attention)
        self.assign(name="encode_decode_attention", component=self.encode_decode_attention)
        self.assign(name="feedforward", component=self.feedforward)

    def pipeline(self,
                 input_data: tf.Tensor,
                 encoder_output: tf.Tensor,
                 encoder_padding_mask: tf.Tensor,
                 decoder_padding_mask: tf.Tensor) -> tf.Tensor:
        output = self.assigned.attention.pipeline(input_data=input_data, mask=decoder_padding_mask)
        output = self.assigned.encode_decode_attention.pipeline(input_data=output,
                                                                mask=encoder_padding_mask,
                                                                encoder_output=encoder_output)
        output = self.assigned.feedforward.pipeline(input_data=output)
        return output


def transformer_encoder_block(obj_name: str,
                              embedding_size: int,
                              head_number: int,
                              scale_rate: int,
                              dropout_rate: float,
                              is_training: Optional[bool]=None) -> Block:
    attention = SelfAttentionLayer(obj_name="attention",
                                   embedding_size=embedding_size,
                                   head_number=head_number,
                                   dropout_rate=dropout_rate,
                                   is_training=is_training)
    feedforward = FeedForwardLayer(obj_name="feedforward",
                                   embedding_size=embedding_size,
                                   scale_rate=scale_rate,
                                   dropout_rate=dropout_rate,
                                   is_training=is_training)
    block = TransformerEncoderBlock(obj_name=obj_name, 
                                    attention=attention, 
                                    feedforward=feedforward)
    return block

def transformer_decoder_block(obj_name: str,
                              embedding_size: int,
                              head_number: int,
                              scale_rate: int,
                              dropout_rate: float,
                              is_training: Optional[bool]=None) -> Block:
    attention = SelfAttentionLayer(obj_name="attention",
                                   embedding_size=embedding_size,
                                   head_number=head_number,
                                   dropout_rate=dropout_rate,
                                   is_training=is_training,
                                   use_look_ahead=True)
    encode_decode_attention = EncoderDecoderAttentionLayer(obj_name="encode_decode_attention",
                                                           embedding_size=embedding_size,
                                                           head_number=head_number,
                                                           dropout_rate=dropout_rate,
                                                           is_training=is_training)
    feedforward = FeedForwardLayer(obj_name="feedforward",
                                   embedding_size=embedding_size,
                                   scale_rate=scale_rate,
                                   dropout_rate=dropout_rate,
                                   is_training=is_training)
    block = TransformerDecoderBlock(obj_name=obj_name, 
                                    attention=attention, 
                                    encode_decode_attention=encode_decode_attention,
                                    feedforward=feedforward)
    return block

from ospark.nn.layers.self_attention import FavorAttentionLayer, EncodeDecodeFavorAttention


def performer_encoder_block(obj_name: str,
                            embedding_size: int,
                            head_number: int,
                            scale_rate: int,
                            dropout_rate: float,
                            random_projections_number: int,
                            is_training: Optional[bool]=None) -> Block:
    attention = FavorAttentionLayer(obj_name="favor_attention",
                                    embedding_size=embedding_size,
                                    head_number=head_number,
                                    dropout_rate=dropout_rate,
                                    is_training=is_training,
                                    random_projections_number=random_projections_number)
    feedforward = FeedForwardLayer(obj_name="feedforward",
                                   embedding_size=embedding_size,
                                   scale_rate=scale_rate,
                                   dropout_rate=dropout_rate,
                                   is_training=is_training)
    block = TransformerEncoderBlock(obj_name=obj_name,
                                    attention=attention,
                                    feedforward=feedforward)
    return block


def performer_decoder_block(obj_name: str,
                            embedding_size: int,
                            head_number: int,
                            scale_rate: int,
                            dropout_rate: float,
                            random_projections_number: int,
                            is_training: Optional[bool]=None) -> Block:
    attention = FavorAttentionLayer(obj_name="favor_attention",
                                    embedding_size=embedding_size,
                                    head_number=head_number,
                                    dropout_rate=dropout_rate,
                                    is_training=is_training,
                                    use_look_ahead=True,
                                    random_projections_number=random_projections_number)
    encode_decode_attention = EncodeDecodeFavorAttention(obj_name="encode_decode_favor_attention",
                                                         embedding_size=embedding_size,
                                                         head_number=head_number,
                                                         dropout_rate=dropout_rate,
                                                         is_training=is_training,
                                                         random_projections_number=random_projections_number)
    feedforward = FeedForwardLayer(obj_name="feedforward",
                                   embedding_size=embedding_size,
                                   scale_rate=scale_rate,
                                   dropout_rate=dropout_rate,
                                   is_training=is_training)
    block = TransformerDecoderBlock(obj_name=obj_name,
                                    attention=attention,
                                    encode_decode_attention=encode_decode_attention,
                                    feedforward=feedforward)
    return block


from ospark import Model, weight_initializer, Layer, Block
from typing import Optional, List, Tuple
from ospark.nn.layers.self_attention import EncoderDecoderAttentionLayer, SelfAttentionLayer
from ospark.nn.layers.gated_linear_units import SwiGLU
from ospark.nn.layers.normalization import LayerNormalization
from ospark.nn.layers.feed_forward import FeedForwardLayer
from ospark.nn.layers.dense_layer import DenseLayer
from ospark.nn.block.transformer_block import TransformerDecoderBlock, TransformerEncoderBlock
from functools import reduce
import tensorflow as tf


class CoCa(Model):

    def __init__(self,
                 obj_name: str,
                 image_encoder: Block,
                 head_number: int,
                 embedding_size: int,
                 embedding_layer: Layer,
                 corpus_size: int,
                 pad_id: Optional[int]=None,
                 scale_rate: Optional[int]=None,
                 pooler: Optional[Layer]=None,
                 unimodal_decoder: Optional[List[Block]]=None,
                 mutimodal_decoder: Optional[List[Block]]=None,
                 mutimodal_block_number: Optional[int]=None,
                 unimodal_block_number: Optional[int]=None,
                 dropout_rate: Optional[float]=None,
                 image_queries_len: Optional[int]=None,
                 trained_weights: Optional[dict]=None,
                 is_training: Optional[bool]=None):
        super(CoCa, self).__init__(obj_name=obj_name, is_training=is_training, trained_weights=trained_weights)

        self._image_encoder     = image_encoder
        self._image_queries_len = image_queries_len or 256
        self._dropout_rate      = dropout_rate or 0.0
        self._embedding_size    = embedding_size
        self._head_number       = head_number
        self._scale_rate        = scale_rate or 4
        self._embedding_layer   = embedding_layer
        self._pad_id            = pad_id or 0
        self._cls_embedding     = weight_initializer.glorot_uniform(obj_name="cls_embedding",
                                                                    shape=[1, 1, self._embedding_size],
                                                                    trainable=is_training)
        self._layer_norm        = LayerNormalization(layer_dimension=embedding_size)

        self._unimodal_block_number  = unimodal_block_number or 2
        self._mutimodal_block_number = mutimodal_block_number or 2

        self._image_queries = weight_initializer.glorot_uniform(obj_name="image_queries",
                                                                shape=[1, self._image_queries_len + 1, embedding_size])

        self._att_pooling = pooler or EncoderDecoderAttentionLayer(obj_name="att_pooling",
                                                                   embedding_size=embedding_size,
                                                                   head_number=head_number,
                                                                   dropout_rate=dropout_rate,
                                                                   is_training=is_training)

        self._unimodal_decoder  = unimodal_decoder or self.create_unimodal(block_number=self._unimodal_block_number)
        self._mutimodal_decoder = mutimodal_decoder or self.create_mutimodal(block_number=self._mutimodal_block_number)

        self._output_layer      = Block(obj_name="logits_layer",
                                        layers=[LayerNormalization(layer_dimension=embedding_size),
                                                DenseLayer(obj_name="dense_layer",
                                                           input_dimension=embedding_size,
                                                           hidden_dimension=[corpus_size],
                                                           use_bias=False)])

    def create_unimodal(self, block_number: int) -> list:
        decoder_blocks = []
        for i in range(block_number):
            att_layer = SelfAttentionLayer(obj_name="att",
                                           embedding_size=self._embedding_size,
                                           head_number=self._head_number,
                                           dropout_rate=self._dropout_rate,
                                           use_look_ahead=True)
            ffnn_layer = FeedForwardLayer(obj_name="ffnn",
                                          embedding_size=self._embedding_size,
                                          scale_rate=self._scale_rate,
                                          dropout_rate=self._dropout_rate,
                                          activation=SwiGLU(dimension=self._embedding_size * self._scale_rate,
                                                            is_training=self.is_training))
            block = TransformerEncoderBlock(obj_name=f"unimodal_block_{i}",
                                            attention=att_layer,
                                            feedforward=ffnn_layer)
            decoder_blocks.append(block)
        return decoder_blocks

    def create_mutimodal(self, block_number: int) -> list:
        decoder_blocks = []
        for i in range(block_number):
            att = SelfAttentionLayer(obj_name="att",
                                     embedding_size=self._embedding_size,
                                     head_number=self._head_number,
                                     dropout_rate=self._dropout_rate,
                                     use_look_ahead=True,
                                     is_training=self.is_training)
            encode_decode_att = EncoderDecoderAttentionLayer(obj_name="encode_decode_att",
                                                             embedding_size=self._embedding_size,
                                                             head_number=self._head_number,
                                                             dropout_rate=self._dropout_rate,
                                                             is_training=self.is_training)

            ffnn_layer = FeedForwardLayer(obj_name="ffnn",
                                          embedding_size=self._embedding_size,
                                          scale_rate=self._scale_rate,
                                          dropout_rate=self._dropout_rate,
                                          activation=SwiGLU(dimension=self._embedding_size * self._scale_rate,
                                                            is_training=self.is_training),
                                          is_training=self.is_training)

            block = TransformerDecoderBlock(obj_name=f"mutimodal{i}",
                                            attention=att,
                                            encode_decode_attention=encode_decode_att,
                                            feedforward=ffnn_layer)
            decoder_blocks.append(block)
        return decoder_blocks

    def pipeline(self, images: tf.Tensor, text: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        batch_size      = len(text)
        image_feature   = self._image_encoder.pipeline(input_data=images)
        image_embedding = self._att_pooling.pipeline(input_data=tf.tile(self._image_queries, [batch_size, 1, 1]),
                                                         encoder_output=image_feature)
        cls_image_embedding, token_image_embedding = image_embedding[:, :1, :], image_embedding[:, 1:, :],

        padding_mask = tf.cast(tf.math.not_equal(text, self._pad_id), tf.float32)[..., tf.newaxis]  # [B, L]
        embeddings   = self._embedding_layer.pipeline(text) * padding_mask

        embeddings = tf.concat([embeddings, tf.tile(self._cls_embedding, [batch_size, 1, 1])], axis=-2)
        mask       = tf.cast(tf.math.equal(embeddings[..., 0], 0), tf.float32)[:, tf.newaxis, :]

        unimodal_output = reduce(lambda input_data, block: block.pipeline(input_data=input_data, mask=mask),
                                 self._unimodal_decoder,
                                 embeddings)

        token_embedding, cls_embedding = unimodal_output[:, :-1, :], unimodal_output[:, -1:, :]

        cls_embedding    = self._layer_norm(input_data=cls_embedding)
        mutimodal_output = reduce(lambda input_data, block: block.pipeline(input_data=input_data,
                                                                           encoder_output=token_image_embedding,
                                                                           encoder_padding_mask=None,
                                                                           decoder_padding_mask=mask[..., :-1]),
                                  self._mutimodal_decoder,
                                  token_embedding)

        logits = tf.nn.softmax(self._output_layer.pipeline(input_data=mutimodal_output))
        return logits, cls_image_embedding, cls_embedding

    def padding(self, input_data: List[tf.Tensor], lengths: List[int]) -> tf.Tensor:
        max_length = max(lengths)
        embeddings = [tf.pad(embedding[tf.newaxis, ...], [[0, 0], [0, max_length - lengths[i]], [0, 0]])
                      for i, embedding in enumerate(input_data)]
        return tf.concat(embeddings, axis=0)


from ospark.nn.optimizer.learning_rate_schedule import TransformerWarmup
from ospark.models.builder import FormerBuilder
from ospark.data.generator.translate_data_generator import TranslateDataGenerator
from ospark.trainer.exdeep_transformer import ExdeepTransformerTrainer
from ospark.nn.loss_function.sparse_categorical_cross_entropy import SparseCategoricalCrossEntropy
from ospark.data.encoder import LanguageDataEncoder
from typing import Optional
import tensorflow_datasets as tfds
import tensorflow_addons as tfa
import json
import os
#os.environ["CUDA_VISIBLE_DEVICES"] = "0"


def text_encoder(folder_path: str,
                 vocab_file: str,
                 vocabulary_size: int,
                 datasets: Optional[list]=None) -> tfds.deprecated.text.SubwordTextEncoder:
    file = os.path.join(folder_path, vocab_file)
    try:
        subword_encoder = tfds.deprecated.text.SubwordTextEncoder.load_from_file(file)
        print(f"載入已建立的字典： {file}")
    except:
        print("沒有已建立的字典，從頭建立。")
        subword_encoder = tfds.deprecated.text.SubwordTextEncoder.build_from_corpus(
            datasets,
            target_vocab_size=vocabulary_size)
        # 將字典檔案存下以方便下次 warm start
        subword_encoder.save_to_file(file)
    return subword_encoder


folder_path          = "/Users/abnertsai/Documents/Gitlab/self-attention/tensorflow_datasets" #
# folder_path          = os.path.dirname(os.path.abspath(__file__))

# dataset_name         = 'wmt14_translate/de-en'
# train_vocab_file     = "de_vocab_file"
# target_vocab_file    = "en_vocab_file"
# train_vocab_file     = "en_vocab_file"
# target_vocab_file    = "de_vocab_file"


dataset_name         = 'ted_hrlr_translate/pt_to_en'
train_vocab_file     = "pt_vocab_file"
target_vocab_file    = "en_vocab_file"

batch_size           = 8
max_length           = None
take_number          = 1000
encoder_block_number = 4
decoder_block_number = 4
head_number          = 4
embedding_size       = 128
scale_rate           = 4
dropout_rate         = 0.0
use_graph_mode       = False
epoch_number         = 50
save_times           = 5
vocabulary_size      = 32000
use_profiling_phase  = False
save_init_weights    = False
use_restore          = True


if use_profiling_phase:
    model_name = "exdeep"
else:
    model_name = "transformer"

# 權重儲存路徑
save_path = os.path.join(folder_path,
            f"{model_name}_{batch_size}_{epoch_number}_{encoder_block_number}_{decoder_block_number}_{head_number}_{embedding_size}.json")

init_weights_path = os.path.join(folder_path,
            f"{encoder_block_number}_{decoder_block_number}_{head_number}_{embedding_size}.json")

# 讀取訓練用的 datasets wmt14
print("讀取訓練用資料")
ds = tfds.load(data_dir=folder_path,
               name=dataset_name,
               as_supervised=True)

train_examples, val_examples = ds["train"], ds["validation"]
print(type(train_examples))
train_datasets, target_datasets = None, None
if not os.path.isfile(os.path.join(folder_path, train_vocab_file + ".subwords")):
    print("找不到 text encoder file，讀取 datasets 建立 text encoder")
    train_datasets, target_datasets = zip(*[[train_data.numpy(), target_data.numpy()]
                                            for train_data, target_data
                                            in train_examples])

# 建立/讀取 text_encoder
train_data_text_encoder = text_encoder(folder_path=folder_path,
                                       vocab_file=train_vocab_file,
                                       vocabulary_size=vocabulary_size,
                                       datasets=train_datasets)

target_data_text_encoder = text_encoder(folder_path=folder_path,
                                        vocab_file=target_vocab_file,
                                        vocabulary_size=vocabulary_size,
                                        datasets=target_datasets)

print("拆分 train data")
training_data, target_data = list(zip(*train_examples))
# 建立 data_generator
print("建立 data generator")
data_encoder   = LanguageDataEncoder(train_data_encoder=train_data_text_encoder,
                                     label_data_encoder=target_data_text_encoder)

data_generator = TranslateDataGenerator(training_data=training_data,
                                        target_data=target_data,
                                        data_encoder=data_encoder,
                                        batch_size=batch_size,
                                        max_length=max_length,
                                        max_token=3000)

if os.path.isfile(init_weights_path) and use_restore:
    print("讀取 init weights")
    with open(init_weights_path, 'r') as fp:
        weights = json.load(fp)
else:
    print("Use random weights")
    weights = None

print("建立模型")
exdeep_model = FormerBuilder.exdeep_transformer(encoder_block_number=encoder_block_number,
                                                decoder_block_number=decoder_block_number,
                                                trained_weights=weights,
                                                head_number=head_number,
                                                embedding_size=embedding_size,
                                                scale_rate=scale_rate,
                                                class_number=target_data_text_encoder.vocab_size + 2,
                                                encoder_corpus_size=train_data_text_encoder.vocab_size + 2,
                                                decoder_corpus_size=target_data_text_encoder.vocab_size + 2,
                                                is_training=True,
                                                dropout_rate=dropout_rate)

# performer_model = FormerBuilder.performer(block_number=encoder_block_number,
#                                           head_number=head_number,
#                                           embedding_size=embedding_size,
#                                           scale_rate=scale_rate,
#                                           random_projections_number=16,
#                                           class_number=target_data_text_encoder.vocab_size + 2,
#                                           encoder_corpus_size=train_data_text_encoder.vocab_size + 2,
#                                           decoder_corpus_size=target_data_text_encoder.vocab_size + 2,
#                                           is_training=True,
#                                           dropout_rate=dropout_rate)

# 設定 learning_rate、optimizer、loss_function
learning_rate = TransformerWarmup(model_dimension=embedding_size, warmup_step=4000.)
# optimizer     = tf.keras.optimizers.Adam(learning_rate=learning_rate, beta_1=0.9, beta_2=0.98, epsilon=1e-9)
optimizer     = tfa.optimizers.RectifiedAdam(learning_rate=learning_rate, beta_1=0.9, beta_2=0.98)
loss_function = SparseCategoricalCrossEntropy()


print("建立 trainer")
trainer = ExdeepTransformerTrainer(data_generator=data_generator,
                                   model=exdeep_model,
                                   epoch_number=epoch_number,
                                   optimizer=optimizer,
                                   loss_function=loss_function,
                                   save_times=save_times,
                                   save_path=save_path,
                                   use_profiling_phase=use_profiling_phase,
                                   use_auto_graph=use_graph_mode,
                                   save_init_weights=save_init_weights,
                                   init_weights_path=init_weights_path)

# trainer = TransformerTrainer(data_generator=data_generator,
#                              model=performer_model,
#                              epoch_number=epoch_number,
#                              optimizer=optimizer,
#                              loss_function=loss_function,
#                              save_times=save_times,
#                              save_path=save_path,
#                              use_auto_graph=use_graph_mode,
#                              save_init_weights=save_init_weights,
#                              init_weights_path=init_weights_path)


print("開始訓練")
trainer.start()


from ospark.nn.layers.normalization import Normalization
from ospark.nn.layers.activation import Activation
from typing import NoReturn, Optional
from . import Layer
import ospark.utility.weight_initializer
import tensorflow as tf
import ospark


class FeedForwardLayer(Layer):

    def __init__(self, 
                 obj_name: str, 
                 embedding_size: int, 
                 scale_rate: int,
                 dropout_rate: float,
                 training_phase: Optional[bool]=None,
                 is_training: Optional[bool]=None,
                 activation: Optional[Activation]=None) -> NoReturn:
        super().__init__(obj_name=obj_name, is_training=is_training, training_phase=training_phase)
        self._activation     = activation or ospark.nn.layers.activation.ReLU()
        self._embedding_size = embedding_size
        self._scale_rate     = scale_rate
        self._dropout_layer  = tf.keras.layers.Dropout(rate=dropout_rate)

    @property
    def embedding_size(self) -> int:
        return self._embedding_size

    @property
    def scale_rate(self) -> int:
        return self._scale_rate

    @property
    def activation(self) -> Activation:
        return self._activation

    @property
    def dropout_layer(self) -> tf.keras.layers.Dropout:
        return self._dropout_layer

    def in_creating(self) -> NoReturn:
        self._mapping2high_dimensional = ospark.utility.weight_initializer.glorot_uniform(
                                obj_name="mapping2high_dimensional",
                                shape=[self.embedding_size, self.scale_rate * self.embedding_size])
        self._mapping2low_dimensional = ospark.utility.weight_initializer.glorot_uniform(
                                obj_name="mapping2low_dimensional",
                                shape=[self.scale_rate * self.embedding_size, self.embedding_size])
        self._high_dimensional_bias = ospark.utility.weight_initializer.zeros(
                                obj_name="high_dimensional_bias",
                                shape=[self.scale_rate * self.embedding_size])
        self._low_dimensional_bias  = ospark.utility.weight_initializer.zeros(
                                obj_name="low_dimensional_bias",
                                shape=[self.embedding_size])

        self._norm = ospark.nn.layers.normalization.LayerNormalization(obj_name="layer_norm",
                                                                       layer_dimension=self.embedding_size)

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        main_output          = self.feedforward(input_data)
        residual_output      = self.residual_net(input_data)
        added_residual       = tf.add(self.dropout_layer(main_output, training=self.training_phase), residual_output)
        normalization_output = self._norm(added_residual)
        return normalization_output

    def feedforward(self, input_data: tf.Tensor) -> tf.Tensor:
        mapping2high_dimensional = tf.matmul(input_data, self._mapping2high_dimensional) + self._high_dimensional_bias
        activated_outputs        = self.activation(mapping2high_dimensional)
        mapping2low_dimensional  = tf.matmul(activated_outputs, self._mapping2low_dimensional) + self._low_dimensional_bias
        return mapping2low_dimensional

    def residual_net(self, input_data: tf.Tensor) -> tf.Tensor:
        return input_data



from ospark.trainer import *
from ospark.nn.model import Model
from ospark.utility.roi_rotate import RoIRotate
from typing import List, Tuple, Optional, Union, Any
from ospark.models.pixel_wise import fots_detection_model, PixelWiseDetection
from ospark.nn.loss_function.loss_function import LossFunction
from ospark.nn.loss_function.degree_loss import Degree
from ospark.nn.loss_function.iou_loss import IoU
from ospark.nn.loss_function.dice_loss import Dice
from ospark.models.text_recognition import TextRecognition, fots_recognition_model
import tensorflow as tf
import numpy as np
import time


class FOTSTrainer(Trainer):

    def __init__(self,
                 data_generator: DataGenerator,
                 detection_model: PixelWiseDetection,
                 recognition_model: TextRecognition,
                 classify_loss_function: LossFunction,
                 bbox_loss_function: LossFunction,
                 degree_loss_function: LossFunction,
                 batch_size: int,
                 epoch_number: int,
                 optimizer: Optimizer,
                 reg_optimizer: Optimizer,
                 corpus: dict,
                 roi_image_height: int=8,
                 recognition_loss_coefficient: Optional[float]=None,
                 save_delegate: Optional[SaveDelegate]=None,
                 save_times: Optional[int]=None,
                 save_path: Optional[str]=None,
                 use_graph: Optional[bool]=True):
        super().__init__(model=detection_model,
                         data_generator=data_generator,
                         epoch_number=epoch_number,
                         optimizer=optimizer,
                         loss_function=None,
                         save_delegate=save_delegate,
                         save_times=save_times,
                         save_path=save_path,
                         use_auto_graph=use_graph,
                         use_multi_gpu=False,
                         devices=None)
        self._roi_rotate                   = RoIRotate(batch_size=batch_size, target_height=roi_image_height)
        self._detection_model              = detection_model
        self._recognition_model            = recognition_model
        self._classify_loss_function       = classify_loss_function
        self._bbox_loss_function           = bbox_loss_function
        self._degree_loss_function         = degree_loss_function
        self._recognition_loss_function    = tf.nn.ctc_loss
        self._corpus                       = corpus
        self._batch_size                   = batch_size
        self._detection_variables          = None
        self._recognition_variables        = None
        self._reg_optimizer                = reg_optimizer
        self._recognition_loss_coefficient = recognition_loss_coefficient or 1.
        self.recognition_model.create()

    @property
    def recognition_model(self) -> Model:
        return self._recognition_model

    @property
    def classify_loss_function(self) -> LossFunction:
        return self._classify_loss_function

    @property
    def bbox_loss_function(self) -> LossFunction:
        return self._bbox_loss_function

    @property
    def degree_loss_function(self) -> LossFunction:
        return self._degree_loss_function

    @property
    def recognition_loss_function(self) -> LossFunction:
        return self._recognition_loss_function

    @property
    def roi_rotate(self) -> RoIRotate:
        return self._roi_rotate

    @property
    def corpus(self) -> dict:
        return self._corpus

    @property
    def batch_size(self) -> int:
        return self._batch_size

    @property
    def detection_variables(self) -> List[tf.Tensor]:
        return self._detection_variables

    @property
    def recognition_variables(self) -> List[tf.Tensor]:
        return self._recognition_variables

    @property
    def recognition_loss_coefficient(self) -> float:
        return self._recognition_loss_coefficient

    @property
    def reg_optimizer(self) -> Optimizer:
        return self._reg_optimizer

    @classmethod
    def create_attention_fots(cls,
                              data_generator: DataGenerator,
                              retrained_weights: dict,
                              corpus: dict,
                              batch_size: int,
                              epoch_number: int,
                              optimizer: Optimizer,
                              reg_optimizer: Optimizer,
                              detection_model: Optional[Model]=None,
                              recognition_model: Optional[Model]=None,
                              save_path: Optional[str]=None,
                              save_times: Optional[int]=None,
                              trainable: Optional[bool]=True):
        detection_model   = detection_model or fots_detection_model(trainable=trainable, retrained_weights=retrained_weights)
        recognition_model = recognition_model or fots_recognition_model(class_number=len(corpus),
                                                                        retrained_weights=retrained_weights,
                                                                        scale_rate=4,
                                                                        head_number=8,
                                                                        input_channel=32,
                                                                        sequential_output_channels=[[64, 64],
                                                                                                    [128, 128],
                                                                                                    [256, 256]],
                                                                        trainable=trainable)
        return cls(data_generator=data_generator,
                   detection_model=detection_model,
                   recognition_model=recognition_model,
                   classify_loss_function=Dice(),
                   bbox_loss_function=IoU(),
                   degree_loss_function=Degree(),
                   batch_size=batch_size,
                   epoch_number=epoch_number,
                   optimizer=optimizer,
                   reg_optimizer=reg_optimizer,
                   corpus=corpus,
                   save_path=save_path,
                   save_times=save_times)

    def start(self) -> NoReturn:
        self._detection_variables   = self.weights_operator.collect_weights("detection_model")
        self._recognition_variables = self.weights_operator.collect_weights("recognition_model")
        for i in range(self.epoch_number):
            start = time.time()
            detection_loss_value   = 0
            recognition_loss_value = 0
            for batch, dataset in enumerate(self.data_generator):
                train_data   = dataset.training_data
                target_maps  = dataset.target_data
                target_words = dataset.words
                bbox_points  = dataset.bbox_points

                feature_maps, detection_loss = self.detection_part(input_data=train_data, target_maps=target_maps)

                sub_images, images_number, target_words, total_width = self.roi_rotate.start(images=feature_maps,
                                                                                             bbox_points=bbox_points,
                                                                                             target_words=target_words)

                recognition_loss = self.recognition_part(sub_images=sub_images,
                                                         target_words=target_words,
                                                         images_number=images_number,
                                                         total_width=total_width)

                detection_loss_value += detection_loss[0] + detection_loss[1] + detection_loss[2]
                recognition_loss_value += recognition_loss
            detection_loss_value   /= batch + 1
            recognition_loss_value /= batch + 1
            print(f"detection loss value: {detection_loss_value:.4f}, "
                  f"recognition loss value: {recognition_loss_value:.4f}")
            print("spent time per epoch: ", time.time() - start)

            if self.will_save(epoch_number=i):
                self.save_delegate.save(weights=self.weights_operator.weights)
        self.save_delegate.save(weights=self.weights_operator.weights)

    @tf.function(input_signature=[
        tf.TensorSpec(shape=[None, None, None, None], dtype=tf.float32),
        tf.TensorSpec(shape=[None, None, None, None], dtype=tf.float32)
    ])
    def detection_part(self,
                       input_data: tf.Tensor,
                       target_maps: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor]:
        with tf.GradientTape() as detection_tape:
            det_predict, feature_maps = self.model.pipeline(input_data)
            loss_value  = self.calculate_detection_loss(prediction=det_predict, target_data=target_maps)
            added_loss  = loss_value[0] + loss_value[1] + loss_value[2]
            gradients   = self.calculate_gradient(tape=detection_tape, loss_value=added_loss, training_variable=self.detection_variables)
        del detection_tape
        self.optimizer.apply_gradients(zip(gradients, self.detection_variables))
        return feature_maps, loss_value

    def calculate_detection_loss(self,
                                 prediction: tf.Tensor,
                                 target_data: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        cls_prediction, bbox_prediction, degree_prediction = prediction

        cls_target, bbox_target, degree_target = self._slice_channel(target_data)

        cls_loss    = self.classify_loss_function(prediction=cls_prediction, target_data=cls_target)
        bbox_loss   = self.bbox_loss_function(prediction=bbox_prediction, target_data=bbox_target, cls_target=cls_target)
        degree_loss = self.degree_loss_function(prediction=degree_prediction, target_data=degree_target, cls_target=cls_target)
        return cls_loss, bbox_loss, degree_loss

    def _slice_channel(self, data) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        cls_channel    = data[:, :, :, 0:1]
        bbox_channel   = data[:, :, :, 1:5]
        degree_channel = data[:, :, :, 5:6]
        return cls_channel, bbox_channel, degree_channel

    def recognition_part(self,
                         sub_images: tf.Tensor,
                         target_words: tf.Tensor,
                         images_number: int,
                         total_width: List[int]) -> Tuple[Union[List[tf.Variable], list], Any]:
        accumulation_gradients        = [tf.Variable(tf.zeros_like(variable.initialized_value()), trainable=False)
                                         for variable in self.recognition_variables]
        total_loss = 0
        counts     = 0
        for images, target_word, images_width in self.sub_images_generator(batch_size=self.batch_size,
                                                                           sub_images=sub_images,
                                                                           target_words=target_words,
                                                                           images_number=images_number,
                                                                           total_width=total_width):

            sparse_target, label_length = self.convert_to_sparse(target_word)
            gradients, loss_value = self.recognition_train_step(training_data=images,
                                                                sparse_target=sparse_target,
                                                                label_length=label_length,
                                                                images_width=images_width)
            total_loss += loss_value
            counts     += 1
            accumulation_gradients = [accumulation_gradients[i].assign_add(var) for i, var in enumerate(gradients)]

        self.reg_optimizer.apply_gradients(zip(accumulation_gradients, self.recognition_variables))
        return total_loss / counts

    @tf.function(input_signature=[
        tf.TensorSpec(shape=[None, None, None, None], dtype=tf.float32),
        tf.SparseTensorSpec(shape=None, dtype=tf.int32),
        tf.TensorSpec(shape=[None], dtype=tf.int32),
        tf.TensorSpec(shape=[None], dtype=tf.int32)
    ])
    def recognition_train_step(self,
                               training_data: tf.Tensor,
                               sparse_target: tf.SparseTensor,
                               label_length: tf.int32,
                               images_width: tf.int32):
        with tf.GradientTape() as recognition_tape:
            training_variable = self.weights_operator.collect_weights("recognition_model")
            recognition_tape.watch(training_variable)
            reg_prediction = self.recognition_model.pipeline(input_data=training_data)
            loss_value  = tf.nn.ctc_loss(labels=sparse_target,
                                         logits=reg_prediction,
                                         label_length=label_length,
                                         logit_length=images_width,
                                         blank_index=0,
                                         logits_time_major=False)
            loss_value  = tf.reduce_mean(loss_value)
            loss_value *= self.recognition_loss_coefficient
            gradients   = self.calculate_gradient(tape=recognition_tape,
                                                  loss_value=loss_value,
                                                  training_variable=training_variable)
        del recognition_tape
        return gradients, loss_value


    def calculate_gradient(self,
                           tape: tf.GradientTape,
                           loss_value: tf.Tensor,
                           training_variable: List[tf.Tensor]):
        tape.watch(training_variable)
        gradient = tape.gradient(loss_value, training_variable)

        gradient = [tf.clip_by_value(grad, -1., 1.) for grad in gradient]
        return gradient

    def sub_images_generator(self,
                             batch_size: int,
                             sub_images: tf.Tensor,
                             target_words: List[List[int]],
                             images_number: int,
                             total_width: List[int]) -> tf.Tensor:
        for i in range(0, images_number, batch_size):
            end_point = min(i + batch_size, images_number)
            yield sub_images[i: end_point, :], target_words[i: end_point], total_width[i: end_point]

    def convert_to_sparse(self, target_sequences: List[List[int]], dtype=np.int32):
        """
        Create a sparse matrix of target_sequences.
        Args:
            target_sequences: List[List[int]]
                a list of lists of type dtype where each element is a sequence
        Returns: Tuple[np.ndarray]
            A tuple with (indices, values, shape)
        """

        indices      = []
        values       = []
        label_length = []
        for n, seq in enumerate(target_sequences):
            seq = [self.corpus[char] for char in seq]
            seq_length = len(seq)
            label_length.append(seq_length)
            indices.extend(zip([n] * len(seq), range(seq_length)))
            values.extend(seq)

        indices = np.asarray(indices, dtype=np.int32)
        values  = np.asarray(values, dtype=dtype)
        shape   = np.asarray([len(target_sequences), indices.max(0)[1] + 1], dtype=np.int32)
        sparse_matrix = tf.SparseTensor(indices=indices, values=values, dense_shape=shape)
        return sparse_matrix, label_length



import ospark
from ospark.nn.layers.convolution_layer import ConvolutionLayer
from ospark.nn.layers import Layer
from typing import NoReturn, Optional
import tensorflow as tf


class ConnectionLayer(Layer):

    def __init__(self,
                 obj_name: str,
                 concatenated_channel: int,
                 output_channel: int,
                 is_training: Optional[bool]=None):
        super().__init__(obj_name=obj_name, is_training=is_training)
        self._concatenated_channel  = concatenated_channel
        self._output_channel        = output_channel
        self._conv_1                = None
        self._conv_2                = None
        self.default_conv()

    @property
    def concatenated_channel(self) -> int:
        return self._concatenated_channel

    @property
    def output_channel(self) -> int:
        return self._output_channel

    @property
    def conv_1(self) -> ConvolutionLayer:
        return self._conv_1

    @property
    def conv_2(self) -> ConvolutionLayer:
        return self._conv_2

    def set_conv_layer(self,
                   conv_1: ConvolutionLayer,
                   conv_2: ConvolutionLayer) -> NoReturn:
        self._conv_1 = conv_1
        self._conv_2 = conv_2

    def default_conv(self) -> NoReturn:
        self._conv_1 = ConvolutionLayer.conv_bn_relu(obj_name="fusion_layer",
                                                     filter_size=[1, 1, self.concatenated_channel, self.output_channel],
                                                     strides=[1, 1, 1, 1],
                                                     padding="SAME",
                                                     is_training=self.is_training)
        self._conv_2 = ConvolutionLayer.conv_bn_relu(obj_name="conv_layer",
                                                     filter_size=[3, 3, self.output_channel, self.output_channel],
                                                     strides=[1, 1, 1, 1],
                                                     padding="SAME",
                                                     is_training=self.is_training)

    def in_creating(self) -> NoReturn:
        self._convolution_1 = self.conv_1
        self._convolution_2 = self.conv_2

    def pipeline(self, input_data: tf.Tensor, connection_input: tf.Tensor) -> tf.Tensor:
        shape          = tf.shape(connection_input)
        height, width  = shape[1], shape[2]
        input_data     = tf.image.resize(input_data , [height, width])
        concat_input   = tf.concat([input_data, connection_input], axis=-1)
        output         = self._convolution_1.pipeline(concat_input)
        output         = self._convolution_2.pipeline(output)
        return output




import tensorflow as tf

from ospark.nn.loss_function.loss_function import LossFunction


class SparseCategoricalCrossEntropy(LossFunction):

    def __init__(self):
        self._loss_function = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True, reduction='none')

    @property
    def loss_function(self) -> tf.keras.losses.Loss:
        return self._loss_function

    def calculate(self, prediction: tf.Tensor, target_data: tf.Tensor) -> tf.Tensor:
        mask  = tf.logical_not(tf.math.equal(target_data, 0))
        loss  = self.loss_function(target_data, prediction)
        mask  = tf.cast(mask, dtype=loss.dtype)
        loss *= mask
        return tf.reduce_sum(loss) / tf.reduce_sum(mask)


from . import *


class CocaTrainer(Trainer):

    def train_step(self, train_data: tf.Tensor, target_data: tf.Tensor):

        with tf.GradientTape() as tape:
            logits, cls_image_embedding, cls_embedding = self.model.pipeline(images=train_data.images,
                                                                             text=train_data.text)
            contrastive_loss = self.loss_function["contrastive_loss"](prediction=cls_image_embedding,
                                                                      target_data=cls_embedding)
            caption_loss = self.loss_function["caption_loss"](prediction=logits,
                                                              target_data=target_data)
            weights    = self.weights_operator.collect_weights()
            tape.watch(weights)
            loss_value = contrastive_loss + caption_loss
        gradients = tape.gradient(loss_value, weights)
        self.optimizer.apply_gradients(zip(gradients, weights))
        return loss_value



from typing import NoReturn


class Metrics:

    def __init__(self) -> NoReturn:
        pass

    def process(self, prediction, target) -> NoReturn:
        return NotImplementedError()

    def calculate_start(self):
        return NotImplementedError()



from typing import NoReturn, Optional
from ospark.nn.component.basic_module import ModelObject
import tensorflow as tf


class Layer(ModelObject):

    def __init__(self, obj_name: str, is_training: Optional[bool]=None, training_phase: Optional[bool]=None) -> NoReturn:
        super().__init__(obj_name=obj_name, is_training=is_training, training_phase=training_phase)

    def in_creating(self) -> NoReturn:
        pass

    def pipeline(self, *args, **kwargs) -> tf.Tensor:
        raise NotImplementedError()


from typing import Tuple, List, Optional
from scipy import sparse
import numpy as np
import tensorflow as tf
import copy
import cv2


class NonMaxSuppression:

    def __init__(self,
                 iou_threshold: Optional[float]=0.5):
        self._iou_threshold = iou_threshold

    @property
    def iou_threshold(self) -> float:
        return self._iou_threshold

    def start(self, bboxes_datasets: List[Tuple[tf.Tensor, tf.Tensor]], image_size: List[int]):
        block_image = np.zeros(shape=image_size, dtype=np.uint8)[:, :, np.newaxis]

        datasets = [
            (score,
             sparse.coo_matrix(np.squeeze(cv2.fillPoly(copy.copy(block_image), [bbox_coordinate], 1))),
             bbox_coordinate) for score, bbox_coordinate in bboxes_datasets
        ]
        datasets = sorted(datasets, key=lambda x:-x[0])
        surviving_bbox = []

        while datasets != []:
            observed_score, observed_polygon, bbox_coordinate = datasets.pop(0)
            datasets_len = len(datasets)

            for i in range(datasets_len):
                reference_score, reference_polygon, reference_bbox_coordinate = datasets.pop(0)
                iou = self.calculate_iou(observed_polygon=observed_polygon, reference_polygon=reference_polygon)
                if iou < self.iou_threshold:
                    datasets.append((reference_score, reference_polygon, reference_bbox_coordinate))
            surviving_bbox.append(bbox_coordinate)
        return surviving_bbox

    def calculate_iou(self,  observed_polygon: sparse.coo_matrix, reference_polygon: sparse.coo_matrix):
        added_polygon = observed_polygon + reference_polygon
        intersection   = len((added_polygon == 2).indices)
        union          = len(added_polygon.nonzero()[0])
        iou = intersection / (union + 1e-5)
        return iou


import math
from ospark.backbone.backbone import Backbone
from ospark.nn.layers import Layer
from typing import *
from ospark.backbone.u_net import Unet
from ospark.nn.layers.convolution_layer import ConvolutionLayer
import tensorflow as tf
from ospark import Model


class PixelWiseDetection(Model):

    def __init__(self,
                 obj_name: str,
                 backbone: Backbone,
                 classify_layer: Layer,
                 trained_weights: Optional[dict]=None):
        super().__init__(obj_name=obj_name, trained_weights=trained_weights)
        self._classify_layer = classify_layer
        self._backbone       = backbone

    @property
    def backbone(self) -> Backbone:
        return self._backbone

    @property
    def classify_layer(self) -> Layer:
        return self._classify_layer

    def pipeline(self, input_data: tf.Tensor) -> Tuple[Tuple[tf.Tensor, tf.Tensor, tf.Tensor], tf.Tensor]:
        feature_map = self.backbone.pipeline(input_data)
        prediction  = self.classify_layer.pipeline(feature_map)

        score_map, bbox_map, angle_map = prediction[:, :, :, 0: 1], prediction[:, :, :, 1: 5], prediction[:, :, :, 5: 6]

        score_map = tf.nn.sigmoid(score_map)
        bbox_map  = tf.nn.sigmoid(bbox_map) * 224
        angle_map = (tf.nn.sigmoid(angle_map) - .5) * math.pi/2
        return (score_map, bbox_map, angle_map), feature_map

    @tf.function
    def __ceil__(self, input_data: tf.Tensor) -> tf.Tensor:
        prediction, feature_map = self.pipeline(input_data=input_data)
        return prediction, feature_map


def fots_detection_model(trainable: bool, retrained_weights):
    backbone = Unet.build_shared_conv(is_training=trainable)
    classify_layer = ConvolutionLayer(obj_name="classify_layer",
                                      filter_size=[1, 1, 32, 6],
                                      strides=[1, 1, 1, 1],
                                      padding="SAME")
    model = PixelWiseDetection(obj_name="detection_model",
                               trained_weights=retrained_weights,
                               backbone=backbone,
                               classify_layer=classify_layer)
    return model


from typing import Tuple

import tensorflow as tf

from ospark.nn.loss_function.loss_function import LossFunction


class IoU(LossFunction):

    def calculate(self,
                  prediction: tf.Tensor,
                  target_data: tf.Tensor) -> tf.Tensor:
        cls_target = tf.cast(tf.where(target_data != 0), tf.float32)
        iou  = self.calculate_iou(prediction, target_data, cls_target)
        loss = -tf.math.log((tf.reduce_sum(iou, axis=[1, 2, 3]) + 1) / (tf.reduce_sum(cls_target, axis=[1, 2, 3]) + 1))
        return tf.reduce_mean(loss)

    def calculate_iou(self,
                      prediction: tf.Tensor,
                      target_data: tf.Tensor,
                      cls_target: tf.Tensor) -> tf.Tensor:
        top_prediction, right_prediction, bottom_prediction, left_prediction = self._slice_channel(prediction)

        top_target, right_target, bottom_target, left_target = self._slice_channel(target_data)

        training_area  = tf.multiply(tf.add(top_prediction, bottom_prediction),
                                     tf.add(right_prediction, left_prediction))
        target_area    = tf.multiply(tf.add(top_target, bottom_target), tf.add(right_target, left_target))
        intersection_h = tf.add(tf.minimum(top_prediction, top_target), tf.minimum(bottom_prediction, bottom_target))
        intersection_w = tf.add(tf.minimum(right_prediction, right_target), tf.minimum(left_prediction, left_target))
        intersection   = tf.multiply(intersection_h, intersection_w)

        union = tf.subtract(tf.add(training_area, target_area), intersection)
        iou   = tf.maximum(tf.divide(intersection + 1, tf.add(union, 1)), 0)
        return iou * cls_target

    def _slice_channel(self, input_data: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor]:
        top    = input_data[:, :, :, 0:1]
        right  = input_data[:, :, :, 1:2]
        bottom = input_data[:, :, :, 2:3]
        left   = input_data[:, :, :, 3:4]
        return top, right, bottom, left

    def __call__(self,
                 prediction: tf.Tensor,
                 target_data: tf.Tensor) -> tf.Tensor:
        loss_value = self.calculate(prediction=prediction, target_data=target_data)
        return loss_value


from ospark.predictor import *
from ospark.models.former import Former
from typing import List
import tensorflow_datasets as tfds
import tensorflow as tf
import numpy as np


class Translator(Predictor):

    def __init__(self,
                 model: Former,
                 input_text_encoder: tfds.deprecated.text.SubwordTextEncoder,
                 output_text_encoder: tfds.deprecated.text.SubwordTextEncoder,
                 predict_max_length: int):
        super().__init__(model=model)
        self._input_text_encoder  = input_text_encoder
        self._output_text_encoder = output_text_encoder
        self._predict_max_length  = predict_max_length
        self._input_bos           = [input_text_encoder.vocab_size]
        self._output_bos          = [output_text_encoder.vocab_size]
        self._input_eos           = [input_text_encoder.vocab_size + 1]
        self._output_eos          = [output_text_encoder.vocab_size + 1]

    @property
    def input_text_encoder(self) -> tfds.deprecated.text.SubwordTextEncoder:
        return self._input_text_encoder

    @property
    def output_text_encoder(self) -> tfds.deprecated.text.SubwordTextEncoder:
        return self._output_text_encoder

    @property
    def predict_max_length(self) -> int:
        return self._predict_max_length

    @property
    def input_bos(self) -> List[int]:
        return self._input_bos

    @property
    def input_eos(self) -> List[int]:
        return self._input_eos

    @property
    def output_bos(self) -> tf.Tensor:
        return tf.convert_to_tensor(self._output_bos, dtype=tf.int32)[tf.newaxis, :]

    @property
    def output_eos(self) -> np.ndarray:
        return np.array(self._output_eos)

    def predict(self, input_data: str):
        input_encoded_text = tf.convert_to_tensor(self.input_bos + \
                                                  self.input_text_encoder.encode(input_data) + \
                                                  self.input_eos, dtype=tf.int32)[tf.newaxis, :]
        decoder_input      = self.output_bos
        for i in range(self.predict_max_length):
            prediction = self.model(encoder_input=input_encoded_text, decoder_input=decoder_input)
            prediction = tf.cast(tf.argmax(prediction, axis=-1), dtype=tf.int32)
            if prediction.numpy()[0, -1] == self.output_eos[0]:
                break
            decoder_input = tf.concat([self.output_bos, prediction], axis=-1)
        text_output = self.output_text_encoder.decode(decoder_input.numpy()[0, 1:])
        return text_output


import ospark.utility.weight_initializer
from ospark.nn.layers.embedding_layer import EmbeddingLayer
from ospark.nn.block import Block
from ospark.nn.component.weight import Weight
from typing import List, Tuple, Optional, Callable, Union
from ospark import Model
import ospark
import numpy as np
import tensorflow as tf


class Former(Model):

    def __init__(self,
                 obj_name: str,
                 encoder_blocks: List[Block],
                 class_number: int,
                 embedding_size: int,
                 dropout_rate: float,
                 trained_weights: Optional[dict]=None,
                 decoder_blocks: Optional[List[Block]]=None,
                 is_training: Optional[bool]=None,
                 encoder_corpus_size: Optional[int]=None,
                 decoder_corpus_size: Optional[int]=None,
                 use_embedding_layer: Optional[bool]=True,
                 use_classifier: Optional[bool]=False,
                 max_length: int=2000
                 ):
        super().__init__(obj_name=obj_name, trained_weights=trained_weights, is_training=is_training)
        self._embedding_size       = embedding_size
        self._max_length           = max_length
        self._encoder_blocks       = encoder_blocks
        self._decoder_blocks       = decoder_blocks or []
        self._class_number         = class_number
        self._embedding_layer      = None
        self._embedding_scale_rate = tf.math.sqrt(tf.cast(self.embedding_size, dtype=tf.float32))
        self._encoder_dropout_layer= tf.keras.layers.Dropout(rate=dropout_rate)
        self._classify_layer       = ospark.utility.weight_initializer.glorot_uniform(obj_name="classify_layer",
                                                                                      shape=[self.embedding_size, class_number])
        self._classify_layer_bias  = ospark.utility.weight_initializer.zeros(obj_name="classify_layer_bias",
                                                                             shape=[class_number])
        self._classifier           = tf.nn.sigmoid if class_number == 2 else tf.nn.softmax
        self._use_classifier       = use_classifier

        self._positional_encoding_table = self.create_positional_encoding_table()

        for component in [*self.encoder_blocks, self.classify_layer, self.classify_layer_bias]:
            self.assign(component)

        if decoder_blocks is not None:
            self._decoder_dropout_layer = tf.keras.layers.Dropout(rate=dropout_rate)
            for component in decoder_blocks:
                self.assign(component)

        if use_embedding_layer:
            if encoder_corpus_size is None:
                raise KeyError("Use embedding layers, must setting encoder_corpus_size")
            self._encoder_embedding_layer = EmbeddingLayer(obj_name="encoder_embedding_layer",
                                                           embedding_dimension=embedding_size,
                                                           corpus_size=encoder_corpus_size)
            self.assign(self.encoder_embedding_layer)

            if decoder_blocks is not None:
                if decoder_corpus_size is None:
                    raise KeyError("Use embedding layers, must setting decoder_corpus_size")
                self._decoder_embedding_layer = EmbeddingLayer(obj_name="decoder_embedding_layer",
                                                               embedding_dimension=embedding_size,
                                                               corpus_size=decoder_corpus_size)
                self.assign(self.decoder_embedding_layer)

            self._create_mask_matrix = self.create_mask_by_onehot

        else:
            self._decoder_embedding_layer = lambda input_data: input_data
            self._encoder_embedding_layer = lambda input_data: input_data
            self._create_mask_matrix      = self.create_mask_by_embedding

    @property
    def embedding_size(self) -> int:
        return self._embedding_size

    @property
    def max_length(self) -> int:
        return self._max_length

    @property
    def positional_encoding_table(self) -> tf.Tensor:
        return self._positional_encoding_table

    @property
    def encoder_blocks(self) -> List[Block]:
        return self._encoder_blocks

    @property
    def decoder_blocks(self) -> List[Block]:
        return self._decoder_blocks

    @property
    def class_number(self) -> int:
        return self._class_number

    @property
    def classify_layer(self) -> Weight:
        return self._classify_layer

    @property
    def classifier(self) -> Callable[[tf.Tensor], tf.Tensor]:
        return self._classifier

    @property
    def use_classifier(self) -> bool:
        return self._use_classifier

    @property
    def encoder_embedding_layer(self) -> Union[EmbeddingLayer, Callable[[tf.Tensor], tf.Tensor]]:
        return self._encoder_embedding_layer

    @property
    def decoder_embedding_layer(self) -> Union[EmbeddingLayer, Callable[[tf.Tensor], tf.Tensor]]:
        return self._decoder_embedding_layer

    @property
    def embedding_scale_rate(self) -> tf.Tensor:
        return self._embedding_scale_rate

    @property
    def classify_layer_bias(self) -> Weight:
        return self._classify_layer_bias

    @property
    def create_mask_matrix(self) -> Callable[[tf.Tensor, Optional[tf.Tensor]], Tuple[tf.Tensor, tf.Tensor]]:
        return self._create_mask_matrix

    @property
    def encoder_dropout_layer(self) -> tf.keras.layers.Dropout:
        return self._encoder_dropout_layer

    @property
    def decoder_dropout_layer(self) -> tf.keras.layers.Dropout:
        return self._decoder_dropout_layer

    def create_mask_by_onehot(self,
                              encoder_input: tf.Tensor,
                              decoder_input: Optional[tf.Tensor]=None) -> Tuple[tf.Tensor, tf.Tensor]:
        encoder_padding_mask = tf.cast(tf.math.equal(encoder_input, 0), tf.float32)[:, tf.newaxis, tf.newaxis, :]
        if decoder_input is not None:
            sequence_length        = tf.shape(decoder_input)[1]
            decoder_padding_mask   = tf.cast(tf.math.equal(decoder_input, 0), tf.float32)[:, tf.newaxis, tf.newaxis, :]
            lookahead_padding_mask = 1 - tf.linalg.band_part(tf.ones((sequence_length, sequence_length)), -1, 0)
            lookahead_padding_mask = tf.maximum(decoder_padding_mask, lookahead_padding_mask)
        else:
            lookahead_padding_mask = None
        return encoder_padding_mask, lookahead_padding_mask

    def create_mask_by_embedding(self,
                                 encoder_input: tf.Tensor,
                                 decoder_input: Optional[tf.Tensor]=None) -> Tuple[tf.Tensor, tf.Tensor]:
        encoder_input = encoder_input[:, :, 0]
        decoder_input = decoder_input[:, :, 0] if decoder_input is not None else None
        encoder_padding_mask, lookahead_padding_mask = self.create_mask_by_onehot(encoder_input=encoder_input,
                                                                                  decoder_input=decoder_input)
        return encoder_padding_mask, lookahead_padding_mask

    def create_positional_encoding_table(self) -> tf.Tensor:
        basic_table = np.zeros(shape=[self.max_length, self.embedding_size])
        position    = np.arange(self.max_length).reshape([-1, 1])
        denominator = np.power(10000, np.arange(0, self.embedding_size, 2) / self.embedding_size)
        basic_table[:, 0::2] = np.sin(position / denominator)
        basic_table[:, 1::2] = np.cos(position / denominator)
        return tf.convert_to_tensor(basic_table, dtype=tf.float32)[tf.newaxis, :, :]

    def pipeline(self, encoder_input: tf.Tensor, decoder_input: Optional[tf.Tensor]=None) -> tf.Tensor:
        raise NotImplementedError()


import ospark.utility.weight_initializer
from ospark.nn.layers import Layer
from typing import List, NoReturn, Optional, Callable, Tuple, Union
from ospark.nn.layers.activation import Activation, PassActivation
from functools import reduce
from ospark import Weight
import tensorflow as tf
import ospark


class DenseLayer(Layer):

    def __init__(self,
                 obj_name: str,
                 input_dimension: int,
                 hidden_dimension: List[int],
                 is_training: Optional[bool]=None,
                 activation: Optional[str]=None,
                 use_bias: Optional[bool]=True):
        super().__init__(obj_name=obj_name, is_training=is_training)
        self._input_dimension  = input_dimension
        self._hidden_dimension = hidden_dimension
        self._layers_name      = []
        self._activation       = PassActivation() if activation is None else getattr(Activation, activation.lower())()
        self._use_bias         = use_bias
        self._forward          = self.bias_forward if use_bias else self.no_bias_forward

    @property
    def input_dimension(self) -> int:
        return self._input_dimension

    @property
    def hidden_dimension(self) -> List[int]:
        return self._hidden_dimension

    @property
    def layers_name(self) -> list:
        return self._layers_name

    @property
    def use_bias(self) -> bool:
        return self._use_bias

    @property
    def forward(self) -> Callable[[tf.Tensor, Union[List[tf.Tensor], tf.Tensor]], tf.Tensor]:
        return self._forward

    def in_creating(self) -> NoReturn:
        for i, output_dimension in enumerate(self.hidden_dimension):
            if i == 0:
                input_dimension = self.input_dimension
            else:
                input_dimension = self.hidden_dimension[i - 1]
            name = f"layer_{i}"
            self._layers_name.append(name)
            setattr(self, name, ospark.utility.weight_initializer.glorot_uniform(obj_name=name, shape=[input_dimension, output_dimension]))
            if self.use_bias:
                setattr(self, name + "_bias", ospark.utility.weight_initializer.zeros(obj_name=name + "_bias", shape=[output_dimension]))

    def bias_forward(self, input_data: tf.Tensor, weight: Tuple[Weight, Weight]) -> tf.Tensor:
        return self._activation(tf.matmul(input_data, weight[0]) + weight[1])

    def no_bias_forward(self, input_data: tf.Tensor, weight: tf.Tensor) -> tf.Tensor:
        return self._activation(tf.matmul(input_data, weight))

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        layers = [(getattr(self, name), getattr(self, name + "_bias"))
                  if self.use_bias else getattr(self, name)
                  for name in self.layers_name]
        output = reduce(self.forward, layers, input_data)
        return output


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ospark',
 'ospark.backbone',
 'ospark.data',
 'ospark.data.generator',
 'ospark.models',
 'ospark.nn',
 'ospark.nn.block',
 'ospark.nn.cell',
 'ospark.nn.component',
 'ospark.nn.layers',
 'ospark.nn.loss_function',
 'ospark.nn.metrics',
 'ospark.nn.optimizer',
 'ospark.ocr',
 'ospark.predictor',
 'ospark.samples',
 'ospark.trainer',
 'ospark.utility',
 'ospark.validator']

package_data = \
{'': ['*']}

install_requires = \
['tensorflow>=2.7.0,<3.0.0']

setup_kwargs = {
    'name': 'ospark',
    'version': '0.1.3',
    'description': 'Ospark is an opensource for quickly builder of former model series.',
    'long_description': None,
    'author': 'ospark-org',
    'author_email': 'donggicai1991@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)



from tensorflow.keras.optimizers.schedules import LearningRateSchedule
import tensorflow as tf 
from enum import Enum

class LearningRateOption(Enum):

    transformer_warmup = lambda model_dimension, warmup_step:TransformerWarmup(model_dimension, warmup_step)


class TransformerWarmup(LearningRateSchedule):

    def __init__(self, model_dimension: int, warmup_step: float=4000.):
        self.model_dimension = tf.cast(model_dimension, dtype=tf.float32)
        self.warmup_step     = warmup_step

    def __call__(self, step):
        learning_rate = 1. / tf.sqrt(self.model_dimension) * \
                        tf.math.minimum(1 / tf.sqrt(step), step * tf.pow(self.warmup_step, -1.5))
        return learning_rate



class PrefixNameMeta(type):

    is_instanced = False
    instance_class = None

    def __call__(mcls, *args, **kwargs):
        if mcls.is_instanced:
            obj = mcls.instance_class
        else:
            obj = super().__call__(*args, **kwargs)
            mcls.instance_class = obj
            mcls.is_instanced = True
        return obj


class PrefixName(metaclass=PrefixNameMeta):

    _prefix_name = ""

    def __add__(self, name: str):
        self._prefix_name += name
        return self

    def __getitem__(self, key):
        self._prefix_name = self._prefix_name[key]
        return self

    @property
    def name(self):
        return self._prefix_name


class NameSpace:

    def __init__(self, name: str):
        self._name   = name
        self._prefix = PrefixName()

    def __enter__(self):
        self._prefix + self._name + "/"

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._prefix = self._prefix[:-(len(self._name) + 1)]


from ospark.nn.layers import Layer, NoReturn
from typing import List
import tensorflow as tf


class PoolingLayer(Layer):

    def __init_subclass__(cls) -> NoReturn:
        super().__init_subclass__()
        setattr(PoolingLayer, cls.__name__, cls)

    def __init__(self,
                 obj_name: str,
                 pooling_size: List[int],
                 strides: List[int],
                 padding: str):
        super().__init__(obj_name=obj_name)
        self._pooling_size = pooling_size
        self._strides      = strides
        self._padding      = padding

    @property
    def pooling_size(self) -> List[int]:
        return self._pooling_size

    @property
    def padding(self) -> str:
        return self._padding

    @property
    def strides(self) -> List[int]:
        return self._strides

    def pipeline(self, input_data: tf.Tensor):
        raise NotImplementedError()

    def __call__(self, *args, **kwargs):
        return self.pipeline(*args, **kwargs)


class Max2DPooling(PoolingLayer):

    def __init__(self,
                 pooling_size: List[int],
                 strides: List[int],
                 padding: str):
        super().__init__(obj_name="max_2d",
                         pooling_size=pooling_size,
                         strides=strides,
                         padding=padding)

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        pooling_result = tf.nn.max_pool2d(input=input_data,
                                          ksize=self.pooling_size,
                                          strides=self.strides,
                                          padding=self.padding)
        return pooling_result


class Average2DPooling(PoolingLayer):

    def __init__(self,
                 pooling_size: List[int],
                 strides: List[int],
                 padding: str):
        super().__init__(obj_name="average_2d",
                         pooling_size=pooling_size,
                         strides=strides,
                         padding=padding)

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        pooling_result = tf.nn.avg_pool2d(input=input_data,
                                          ksize=self.pooling_size,
                                          strides=self.strides,
                                          padding=self.padding)
        return pooling_result


from __future__ import annotations
from ospark.nn.block import Block
from ospark.nn.layers import Layer
from ospark.nn.layers.normalization import Normalization, BatchNormalization
from ospark.nn.layers.activation import Activation, ReLU
from ospark.nn.layers.convolution_layer import ConvolutionLayer
from typing import List, NoReturn, Optional
from functools import reduce
import tensorflow as tf


class VGGBlock(Block):

    def __init__(self,
                 obj_name: str,
                 layers: List[Layer],
                 pooling_size: List[int],
                 pooling_strides: List[int]):
        super().__init__(obj_name=obj_name)
        self._layers          = layers
        self._pooling_size    = pooling_size
        self._pooling_strides = pooling_strides

    @property
    def layers(self) -> List[Layer]:
        return self._layers

    @property
    def pooling_size(self) -> List[int]:
        return self._pooling_size

    @property
    def pooling_sreides(self) -> List[int]:
        return self._pooling_strides

    @classmethod
    def create_vgg_block(cls,
                         obj_name: str,
                         input_channel: int,
                         output_channels: List[int],
                         pooling_size: List[int],
                         pooling_strides: List[int],
                         activation: List[Activation],
                         normalization: List[Normalization],
                         filter_size: Optional[List[int]]=[3, 3],
                         filter_strides: Optional[List[int]]=[1, 1, 1, 1],
                         layer_order: Optional[List[str]]=["conv", "norm", "activate"]) -> VGGBlock:
        layers = []
        for i, output_channel in enumerate(output_channels):
            if i == 0:
                input_channel = input_channel
            else:
                input_channel = output_channels[i - 1]

            layer_name = f"layer_{i}"
            conv_layer = ConvolutionLayer(obj_name=layer_name,
                                          filter_size=[*filter_size, input_channel, output_channel],
                                          strides=filter_strides,
                                          padding="SAME",
                                          activation=activation[i],
                                          normalization=normalization[i],
                                          layer_order=layer_order)
            layers.append(conv_layer)
        return cls(obj_name=obj_name,
                   layers=layers,
                   pooling_size=pooling_size,
                   pooling_strides=pooling_strides)


    def in_creating(self) -> NoReturn:
        for layer in self.layers:
            self.assign(component=layer)

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        layer_output = reduce(lambda output, layer: layer.pipeline(output), self.layers, input_data)
        block_output = tf.nn.max_pool2d(input=layer_output,
                                        strides=self.pooling_sreides,
                                        ksize=self.pooling_size,
                                        padding="SAME")
        return block_output

def fots_like_vgg(input_channel: int,
                  output_channels: List[List[int]],
                  trainable: Optional[bool]=True
                  ) -> List[VGGBlock]:
    """
    The vgg model used by FOTS
    
    Args:
        input_channel: int
            Input channel of vgg model, type is int32
        output_channels:  List[List[int]]
            2-D structure [layer_number, output_channels],
    Returns:

    """

    vgg_blocks = []
    for i, layer_channels in enumerate(output_channels):
        if i == 0:
            input_channel = input_channel
        else:
            input_channel = output_channels[i - 1][-1]
        name      = f"vgg_block_{i}"
        vgg_block = VGGBlock.create_vgg_block(obj_name=name,
                                              input_channel=input_channel,
                                              output_channels=layer_channels,
                                              pooling_size=[1, 2, 1, 1],
                                              pooling_strides=[1, 2, 1, 1],
                                              normalization=[BatchNormalization(obj_name="batch_norm",
                                                                                input_depth=channel,
                                                                                is_training=trainable) for channel in layer_channels],
                                              activation=[ReLU() for i in layer_channels])
        vgg_blocks.append(vgg_block)
    return vgg_blocks






from ospark.nn.component.weight import WeightOperator
from ospark.data.generator import DataGenerator
from ospark.data.save_delege import SaveDelegate
from typing import NoReturn, Optional, Callable, List, Tuple, Union, Dict
from tensorflow.keras.optimizers import Optimizer
from ospark.nn.loss_function.loss_function import LossFunction
from ospark.nn.model import Model
import tensorflow as tf
import json
import time


class Trainer:

    def __init__(self,
                 model: Model,
                 data_generator: DataGenerator,
                 epoch_number: int,
                 optimizer: Optimizer,
                 loss_function: Union[LossFunction, Dict[str, LossFunction]],
                 save_delegate: Optional[SaveDelegate]=None,
                 save_times: Optional[int]=None,
                 save_path: Optional[str]=None,
                 use_auto_graph: Optional[bool]=True,
                 use_multi_gpu: Optional[bool]=None,
                 devices: Optional[List[str]]=None):
        self._weights_operator = WeightOperator()
        self._loss_function    = loss_function
        self._epoch_number     = epoch_number
        self._save_delegate    = save_delegate or self
        self._save_times       = save_times
        self._save_path        = save_path

        use_multi_gpu = use_multi_gpu or False
        if use_multi_gpu:
            self._mirrored_strategy = tf.distribute.MirroredStrategy() if devices is None \
                                      else tf.distribute.MirroredStrategy(devices=devices)

            with self.mirrored_strategy.scope():
                self._model     = model
                self._optimizer = optimizer

            self._data_generator  = self.mirrored_strategy.experimental_distribute_dataset(data_generator)
            self._training_method = self.distributed_train_step

        else:
            self._model            = model
            self._optimizer        = optimizer
            self._data_generator   = data_generator
            self._training_method  = self.graph_mode if use_auto_graph else self.eager_mode

    @property
    def model(self) -> Model:
        return self._model

    @property
    def loss_function(self) -> LossFunction:
        return self._loss_function

    @property
    def mirrored_strategy(self) -> tf.distribute.MirroredStrategy:
        return self._mirrored_strategy

    @property
    def weights_operator(self) -> WeightOperator:
        return self._weights_operator

    @property
    def data_generator(self) -> DataGenerator:
        return self._data_generator

    @property
    def epoch_number(self) -> int:
        return self._epoch_number

    @property
    def optimizer(self) -> Optimizer:
        return self._optimizer

    @property
    def save_delegate(self) -> SaveDelegate:
        return self._save_delegate

    @property
    def save_times(self) -> int:
        return self._save_times

    @property
    def save_path(self) -> str:
        return self._save_path

    @property
    def training_method(self) -> Callable:
        return self._training_method

    def start(self):
        print("=" * 24)
        print("Training start.")
        self.training_process()
        print("Training end.")
        print("=" * 24)

    def training_process(self) -> NoReturn:
        for epoch in range(self.epoch_number):
            total_loss_value = 0
            training_count   = 0
            start_time       = time.time()
            for batch, dataset in enumerate(self.data_generator):
                training_data, target_data = dataset.training_data, dataset.target_data
                loss_value = self.training_method(training_data, target_data)
                total_loss_value += loss_value
                training_count   += 1

            print(f'Epoch {epoch + 1} '
                  f'Loss {total_loss_value / training_count:.4f} ')
            print(f'Time taken for 1 epoch: {time.time() - start_time:.2f} secs\n')
            if self.will_save(epoch_number=epoch) and self.save_path is not None:
                self.save_delegate.save(weights=self.weights_operator.weights, path=self.save_path)

        if self.save_path is not None:
            self.save_delegate.save(weights=self.weights_operator.weights, path=self.save_path)

    def train_step(self, train_data: tf.Tensor, target_data: tf.Tensor):
        with tf.GradientTape() as tape:
            prediction = self.model.pipeline(input_data=train_data)
            loss_value = self.loss_function(prediction=prediction, target_data=target_data)
            weights    = self.weights_operator.collect_weights()
            tape.watch(weights)
        gradients = tape.gradient(loss_value, weights)
        self.optimizer.apply_gradients(zip(gradients, weights))
        return loss_value

    @tf.function(experimental_relax_shapes=True)
    def graph_mode(self,
                   train_data: tf.Tensor,
                   target_data: tf.Tensor):
        return self.train_step(train_data=train_data, target_data=target_data)

    def eager_mode(self,
                   train_data: tf.Tensor,
                   target_data: tf.Tensor):
        return self.train_step(train_data=train_data, target_data=target_data)

    def get_weights(self) -> dict:
        return self.weights_operator.weights

    def will_save(self, epoch_number: int) -> bool:
        if self.save_times is None:
            return False
        return (epoch_number + 1) % self.save_times == 0

    @tf.function(experimental_relax_shapes=True)
    def distributed_train_step(self,
                               train_data: tf.Tensor,
                               target_data: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor]:
        accuracies, per_replica_losses = self.mirrored_strategy.run(self.train_step, args=(train_data, target_data))
        return (self.mirrored_strategy.reduce(tf.distribute.ReduceOp.MEAN, accuracies, axis=None),
                self.mirrored_strategy.reduce(tf.distribute.ReduceOp.MEAN, per_replica_losses, axis=None))

    def save(self, weights: dict, path: str) -> NoReturn:
        with open(path, 'w') as fp:
            json.dump(weights, fp)





import json
from ospark.nn.model import Model
from  pathlib import Path


class DataOperator:

    @staticmethod
    def load(path: str) -> dict:
        raise NotImplementedError()

    @staticmethod
    def save(path: str, obj: dict) -> None:
        raise NotImplementedError()


class JsonOperator(DataOperator):

    @staticmethod
    def load(path: str) -> dict:
        with open(path, 'r') as fp:
            json_data = json.load(fp=fp)
        return json_data

    @staticmethod
    def save(folder_path: str, model: Model) -> None:
        save_path = Path(folder_path) / (model.obj_name + "_weights.json")
        with open(save_path, 'w') as fp:
            json.dump(obj=model.get_weights(), fp=fp)


from ospark.nn.cell.resnet_cell import ResnetCell
from ospark.nn.layers.convolution_layer import ConvolutionLayer
from typing import List, NoReturn, Optional
from ospark.backbone.backbone import Backbone
import tensorflow as tf


class ResnetBackbone(Backbone):

    def __init__(self,
                 obj_name: str,
                 cells: List[ResnetCell],
                 is_keep_blocks: bool,
                 is_training: Optional[bool]=None):
        super().__init__(obj_name=obj_name,
                         is_keep_blocks=is_keep_blocks,
                         is_training=is_training)
        self._cells = cells

    @property
    def cells(self) -> List[ResnetCell]:
        return self._cells

    def in_creating(self) -> NoReturn:
        self.assign(component=ConvolutionLayer.conv_bn_relu(obj_name="first_conv",
                                                            filter_size=[7, 7, 3, 64],
                                                            strides=[1, 2, 2, 1],
                                                            padding="SAME",
                                                            is_training=self.is_training))
        for cell in self.cells:
            self.assign(component=cell)

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        output = self.assigned.first_conv.pipeline(input_data)
        output = tf.nn.max_pool2d(output, strides=2, ksize=3, padding="SAME")
        for cell in self.cells:
            output = cell.pipeline(output)
            if self.is_keep_blocks:
                self._block_register.append(output)
        output = self.block_register if self.is_keep_blocks else output
        return output


from __future__ import annotations
from ospark.nn.component.basic_module import ModelObject
from typing import List, NoReturn, Union, Optional
from ospark.nn.layers import Layer
from abc import abstractmethod
import tensorflow as tf 

class Block(ModelObject):

    def __init__(self,
                 obj_name: str,
                 layers: Optional[List[ModelObject]]=None,
                 is_training: Optional[bool]=None,
                 training_phase: Optional[bool]=None):
        super().__init__(obj_name=obj_name, is_training=is_training, training_phase=training_phase)
        self._layers = layers or []

    @property
    def layers(self) -> List[Layer]:
        return self._layers

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        output = input_data
        for layer in self.layers:
            output = layer.pipeline(input_data=output)
        return output

    def in_creating(self) -> NoReturn:
        for layer in self.layers:
            self.assign(component=layer)

    @classmethod
    def draw_structure(cls, layer_classes: List[Layer]):
        def set_configs(obj_name: str, configs: list) -> Block:
            block = cls(obj_name=obj_name)
            for layer, config in zip(layer_classes, configs):
                block.assign(layer(**config))
            return block
        return set_configs

    def add_layers(self, layers: Union[ModelObject, List[ModelObject]]):
        if isinstance(layers, ModelObject):
            self._layers.append(layers)
        elif isinstance(layers, list):
            self._layers += layers
        else:
            raise TypeError("Input type must be ModelObject or List[ModelObject]")


from __future__ import annotations
from typing import NoReturn, Optional, Type
from ospark.nn.layers.self_attention import SelfAttentionLayer, EncoderDecoderAttentionLayer
from ospark.nn.layers.prob_sparse import ProbSparseAttentionLayer
from ospark.nn.layers.distilling_layer import DistillingLayer
from ospark.nn.layers.feed_forward import FeedForwardLayer
from ospark.nn.layers.activation import GELU
from . import Block
import tensorflow as tf

class InformerEncoderBlock(Block):

    def __init__(self,
                 obj_name: str,
                 attention: SelfAttentionLayer,
                 feedforward: FeedForwardLayer,
                 distilling: DistillingLayer) -> NoReturn:
        super().__init__(obj_name)
        self._attention      = attention
        self._feedforward    = feedforward
        self._distilling     = distilling
 
    @property
    def attention(self) -> SelfAttentionLayer:
        return self._attention

    @property
    def feedforward(self) -> FeedForwardLayer:
        return self._feedforward

    @property
    def distilling(self) -> DistillingLayer:
        return self._distilling

    @classmethod
    def create_via_class(cls,
                         obj_name: str,
                         embedding_size: int,
                         head_number: int,
                         scale_rate: int,
                         attention_cls: Type[SelfAttentionLayer],
                         feedforward_cls: Type[FeedForwardLayer],
                         distilling_cls: Type[DistillingLayer],
                         is_training: Optional[bool]=None) -> InformerEncoderBlock:
        return cls(obj_name=obj_name,
                   attention=attention_cls(obj_name="attention",
                                           embedding_size=embedding_size,
                                           head_number=head_number,
                                           is_training=is_training),
                   feedforward=feedforward_cls(obj_name="feedforward",
                                               embedding_size=embedding_size,
                                               scale_rate=scale_rate,
                                               is_training=is_training),
                   distilling=distilling_cls(obj_name="distilling",
                                             embedding_size=embedding_size,
                                             is_training=is_training))

    def in_creating(self) -> NoReturn:
        self.assign(component=self.attention, name="attention")
        self.assign(component=self.feedforward, name="feedforward")
        self.assign(component=self.distilling, name="distilling")

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        layers = [self.assigned.attention(None),
                  self.assigned.feedforward,
                  self.assigned.distilling]
        output = input_data
        for layer in layers:
            output = layer.pipeline(output)
        return output


class InformerDecoderBlock(Block):

    def __init__(self,
                 obj_name: str,
                 attention: SelfAttentionLayer,
                 encode_decode_attention: EncoderDecoderAttentionLayer,
                 feedforward: FeedForwardLayer) -> NoReturn:
        super().__init__(obj_name)
        self._attention               = attention
        self._encode_decode_attention = encode_decode_attention
        self._feedforward             = feedforward

    @property
    def attention(self) -> SelfAttentionLayer:
        return self._attention

    @property
    def encode_decode_attention(self) -> EncoderDecoderAttentionLayer:
        return self._encode_decode_attention

    @property
    def feedforward(self) -> FeedForwardLayer:
        return self._feedforward

    def in_creating(self) -> NoReturn:
        self.assign(component=self.attention, name="attention")
        self.assign(component=self.encode_decode_attention, name="encode_decode_attention")
        self.assign(component=self.feedforward, name="feedforward")

    def pipeline(self, input_data: tf.Tensor, encoder_output: tf.Tensor) -> tf.Tensor:
        layers = [self.assigned.attention(None),
                  self.assigned.encode_decode_attention(mask=None, encoder_output=encoder_output),
                  self.feedforward]
        output = input_data
        for layer in layers:
            output = layer(output)
        return output


def informer_encoder_block(obj_name: str,
                           embedding_size: int,
                           head_number: int,
                           scale_rate: int,
                           sample_factor: float,
                           dropout_rate: float,
                           is_training: Optional[bool]=None,
                           filter_width: int=None,
                           pooling_size: list=None,
                           strides: list=None) -> Block:
    attention   = ProbSparseAttentionLayer("attention",
                                           embedding_size=embedding_size,
                                           head_number=head_number,
                                           sample_factor=sample_factor,
                                           dropout_rate=dropout_rate,
                                           is_training=is_training)
    feedforward = FeedForwardLayer("feedforward",
                                   embedding_size=embedding_size,
                                   scale_rate=scale_rate,
                                   activation=GELU(),
                                   dropout_rate=dropout_rate,
                                   is_training=is_training)
    distilling  = DistillingLayer("distilling",
                                  embedding_size=embedding_size,
                                  filter_width=filter_width,
                                  pooling_size=pooling_size,
                                  strides=strides,
                                  is_training=is_training)
    block = InformerEncoderBlock(obj_name=obj_name,
                                 attention=attention,
                                 feedforward=feedforward,
                                 distilling=distilling)
    return block

def informer_decoder_block(obj_name: str,
                           embedding_size: int,
                           head_number: int,
                           scale_rate: int,
                           sample_factor: float,
                           dropout_rate: float,
                           is_training: Optional[bool]=False) -> Block:
    attention = ProbSparseAttentionLayer("sparse_attention",
                                         embedding_size=embedding_size,
                                         head_number=head_number,
                                         sample_factor=sample_factor,
                                         use_look_ahead=True,
                                         dropout_rate=dropout_rate,
                                         is_training=is_training)
    encode_decode_attention = EncoderDecoderAttentionLayer("encode_decode_attention",
                                                           embedding_size=embedding_size,
                                                           head_number=head_number,
                                                           dropout_rate=dropout_rate,
                                                           is_training=is_training)
    feedforward = FeedForwardLayer("feedforward",
                                   embedding_size=embedding_size,
                                   scale_rate=scale_rate,
                                   activation=GELU(),
                                   dropout_rate=dropout_rate,
                                   is_training=is_training)
    block = InformerDecoderBlock(obj_name=obj_name,
                                 attention=attention,
                                 encode_decode_attention=encode_decode_attention,
                                 feedforward=feedforward)
    return block




from ospark.models.transformer import Transformer
from ospark.models.informer import Informer
from ospark.models.former import Former

def transformer(block_number: int) -> Former:
    return Transformer.quick_build(block_number=block_number)

def exdeep_transformer() -> Former:
    return Transformer.quick_build_exdeep()

def informer() -> Former:
    return Informer.quick_build()


import tensorflow as tf

import ospark.utility.weight_initializer
from . import Layer
from typing import NoReturn, List, Optional, Callable, Union
import ospark


class Normalization(Layer):

    def __init__(self,
                 obj_name: str,
                 layer_dimension: Union[int, list],
                 training_phase: Optional[bool]=None,
                 is_training: Optional[bool]=None,
                 gamma: Optional[float]=None,
                 beta: Optional[float]=None,
                 epsilon: Optional[float]=None,
                 use_bias: Optional[bool]=None,
                 use_scale: Optional[bool]=None):
        super().__init__(obj_name=obj_name, is_training=is_training, training_phase=training_phase)
        self._layer_dimension = layer_dimension

        self._gamma     = gamma or 1.0
        self._beta      = beta or 0.0
        self._epsilon   = epsilon or 0.0001
        self._use_bias  = use_bias or True
        self._use_scale = use_scale or True

    @property
    def gamma(self) -> Union[float, ospark.Weight]:
        return self._gamma

    @property
    def beta(self) -> Union[float, ospark.Weight]:
        return self._beta

    @property
    def use_bias(self) -> bool:
        return self._use_bias

    @property
    def use_scale(self) -> bool:
        return self._use_scale

    @property
    def epsilon(self) -> tf.Tensor:
        return tf.constant(self._epsilon)

    def in_creating(self) -> NoReturn:
        if self.use_scale:
            self._gamma = ospark.utility.weight_initializer.ones(obj_name="gamma", shape=self._layer_dimension) * self._gamma
        if self.use_bias:
            self._beta = ospark.utility.weight_initializer.ones(obj_name="beta", shape=self._layer_dimension) * self._beta

    def __init_subclass__(cls) -> NoReturn:
        super().__init_subclass__()
        setattr(Normalization, cls.__name__, cls)

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        raise NotImplementedError()

    def __call__(self, input_data: tf.Tensor) -> tf.Tensor:
        return self.pipeline(input_data)


class PassNormalization(Normalization):

    def __init__(self):
        super(PassNormalization, self).__init__(obj_name="pass_norm",
                                                layer_dimension=0,
                                                is_training=False)

    def in_creating(self) -> NoReturn:
        pass

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        return input_data


class LayerNormalization(Normalization):

    def __init__(self,
                 layer_dimension: Union[int, list],
                 is_training: Optional[bool]=None,
                 obj_name: Optional[str]=None,
                 gamma: Optional[float]=None,
                 beta: Optional[float]=None,
                 epsilon: Optional[float]=None,
                 use_bias: Optional[bool]=True,
                 use_scale: Optional[bool]=True):
        super().__init__(obj_name=obj_name or "layer_norm",
                         is_training=is_training,
                         layer_dimension=layer_dimension,
                         gamma=gamma,
                         beta=beta,
                         epsilon=epsilon,
                         use_bias=use_bias,
                         use_scale=use_scale)

    def pipeline(self, input_data: tf.Tensor, axis: int=-1) -> tf.Tensor:
        mean, variance = tf.nn.moments(input_data, axes=[axis], keepdims=True)
        normalization_outputs = (input_data - mean) / (tf.sqrt(variance) + self.epsilon)
        if self.use_scale:
            normalization_outputs *= self._gamma
        if self.use_bias:
            normalization_outputs += self._beta
        return normalization_outputs


class BatchNormalization(Normalization):

    def __init__(self,
                 input_depth: Union[tf.Tensor, int],
                 training_phase: Optional[bool]=None,
                 is_training: Optional[bool]=None,
                 obj_name: str=None,
                 gamma: Optional[float]=None,
                 beta: Optional[float]=None,
                 epsilon: Optional[float]=None,
                 momentum: Optional[float]=None,
                 moving_mean: Optional[float]=None,
                 moving_variance: Optional[float]=None,
                 use_bias: Optional[bool]=True,
                 use_scale: Optional[bool]=True):
        super(BatchNormalization, self).__init__(obj_name=obj_name or "batch_norm",
                                                 is_training=is_training,
                                                 layer_dimension=input_depth,
                                                 gamma=gamma,
                                                 beta=beta,
                                                 epsilon=epsilon,
                                                 use_scale=use_scale,
                                                 use_bias=use_bias,
                                                 training_phase=training_phase)
        self._input_depth     = input_depth
        self._momentum        = momentum or 0.9
        self._moving_mean     = moving_mean or 0.0
        self._moving_variance = moving_variance or 1.0

    @property
    def input_depth(self) -> Union[tf.Tensor, int]:
        return self._input_depth

    @property
    def momentum(self) -> tf.Tensor:
        return tf.constant(self._momentum)

    def pipeline(self, input_data: tf.Tensor) -> tf.Tensor:
        if self.training_phase:
            return self.train_process(input_data=input_data)
        else:
            return self.inference_process(input_data=input_data)

    def in_creating(self) -> NoReturn:
        super().in_creating()
        self.assign(component=ospark.utility.weight_initializer.ones(obj_name="moving_mean",
                                                                     shape=[self.input_depth],
                                                                     trainable=False) * self._moving_mean)
        self.assign(component=ospark.utility.weight_initializer.ones(obj_name="moving_variance",
                                                                     shape=[self.input_depth],
                                                                     trainable=False) * self._moving_variance)

    def train_process(self, input_data: tf.Tensor) -> tf.Tensor:
        mean, variance  = tf.nn.moments(x=input_data, axes=[0, 1, 2])
        moving_mean     = self.assigned.moving_mean * self.momentum + mean * (1 - self.momentum)
        self.assigned.moving_mean.assign(moving_mean)
        moving_variance = self.assigned.moving_variance * self.momentum + variance * (1 - self.momentum)
        self.assigned.moving_variance.assign(moving_variance)
        normalization_outputs = tf.nn.batch_normalization(x=input_data,
                                                          mean=mean,
                                                          variance=variance,
                                                          offset=None,
                                                          scale=None,
                                                          variance_epsilon=self.epsilon)
        if self.use_scale:
            normalization_outputs *= self._gamma
        if self.use_bias:
            normalization_outputs += self._beta
        return normalization_outputs

    def inference_process(self, input_data: tf.Tensor) -> tf.Tensor:
        normalization_outputs = tf.nn.batch_normalization(x=input_data,
                                                          mean=self.assigned.moving_mean,
                                                          variance=self.assigned.moving_variance,
                                                          offset=None,
                                                          scale=None,
                                                          variance_epsilon=self.epsilon)
        if self.use_scale:
            normalization_outputs *= self._gamma
        if self.use_bias:
            normalization_outputs += self._beta
        return normalization_outputs


from .loss_function import LossFunction
from typing import Optional
import tensorflow as tf


class ContrastiveLoss(LossFunction):

    def __init__(self, temperature: Optional[float]=None):
        self._temperature = temperature or 1.

    def calculate(self, prediction: tf.Tensor, target_data: tf.Tensor) -> tf.Tensor:
        matmul_output = tf.matmul(prediction, target_data, transpose_b=True)

        part_1_loss = tf.math.log(tf.exp(tf.linalg.diag_part(matmul_output) / self._temperature) / tf.reduce_sum(tf.exp(matmul_output / self._temperature), axis=1))
        part_2_loss =  tf.math.log(tf.exp(tf.linalg.diag_part(matmul_output) / self._temperature) / tf.reduce_sum(tf.exp(tf.transpose(matmul_output) / self._temperature), axis=1))
        loss = -1 * tf.reduce_mean(part_1_loss + part_2_loss)
        return loss